import org.apache.spark.ml.evaluation.RegressionEvaluator
import org.apache.spark.ml.regression.{DecisionTreeRegressionModel,DecisionTreeRegressor}
import org.apache.spark.ml.regression.{RandomForestRegressionModel, RandomForestRegressor}
import org.apache.spark.ml.regression.{GBTRegressionModel, GBTRegressor}
import org.apache.spark.ml.linalg.Vectors
import org.apache.spark.ml.feature.LabeledPoint
import org.apache.spark.rdd.RDD
import org.apache.spark.sql.DataFrame
import scala.collection.mutable.ArrayBuffer
import spark.implicits._


// Número de cada atributo
val ATTR_time = 0
val ATTR_month = 1
val ATTR_hour = 2
val ATTR_demand = 3
val ATTR_pressure = 4
val ATTR_work_day = 5
val ATTR_season = 6
val ATTR_date = 7

val DATA_TRAINING = "training"
val DATA_TEST = "test"

// Nombre de los modelos
val DECISION_TREE = "DecisionTreeRegressor"
val RANDOM_FOREST = "RandomForestRegressor"
val GBT = "GBTRegressor"

// Parametros para los modelos
var MAXBINS = 24
val NUM_TREES = 60

/***************************************************
******************** FUNCIONES *********************
****************************************************/

/* Función que lee los archivos de entrenamiento y prueba y los transforma a RDD */
def readData(): (RDD[Array[String]], RDD[Array[String]]) = {
  // Leemos datos y creamos RDD de String
  var data_training = sc.textFile(DATA_TRAINING)
  var data_test_raw = sc.textFile(DATA_TEST)

  //Dividimos por comas
  var data:RDD[Array[String]] =  data_training.map(line => line.split (";"))
  var dataTest:RDD[Array[String]] =  data_test_raw.map(line => line.split (";"))
  
  (data, dataTest)
}

/* Función que transforma las características a codificación binaria y los atributos numéricos a Double */
def get_features (row: Array[String]): Array[Double] = {
  var attr = 0
  var cat_vec = Array[Double]()
  var num_vec = Array[Double]()

  // Transform the numerical features to double
  for (attr <- 1 to 2){
    cat_vec = cat_vec :+ row(attr).toDouble
  }
  for (attr <- 5 to 6){
    num_vec = num_vec :+ row(attr).toDouble
  }
// Concatenate the two array of features
  cat_vec ++ num_vec
//var attr = 0
}

/* Función que devuelve el RDD de LabeledPoint a partir del RDD de Array de String */
def getDataFrame(data: RDD[Array[String]]): DataFrame = {
    data.map{row => LabeledPoint(
      (row(ATTR_demand).toDouble).toDouble, 
      Vectors.dense(get_features(row)))
    }.toDF("label", "features")
}

/* Función que devuelve la unión de los RDD de un Array de RDD */
def unionRDD(arrayRdd: Array[RDD[Array[String]]]): RDD[Array[String]] = {
    var tam = arrayRdd.length
    var rdd = arrayRdd(0)

    if (tam == 1) arrayRdd
    else {
      var splits = arrayRdd.splitAt(tam/2)
      rdd = unionRDD(splits._1).union(unionRDD(splits._2))
    }
    rdd
}

/* Función que realiza la partición del conjunto en test y train */
def getSubconjuntos (subconjuntos: Array[RDD[Array[String]]], numTest: Int): (RDD[Array[String]], RDD[Array[String]]) = {
    var numElem = subconjuntos.length
    var test: RDD[Array[String]] = subconjuntos(numTest)
    var trainArray: Array[RDD[Array[String]]] = subconjuntos.take(numTest)++subconjuntos.takeRight(numElem-(numTest+1))

    //Concatenamos para crear un único RDD de entrenamiento
    var trainCV = unionRDD(trainArray)
  
   (test,trainCV)
}

/***************************************************
*****************  FUNCIONES EN TEST ***************
****************************************************/

/* Función que devuelve el DataFrame con las predicciones y las labels de un modelo DecisionTreeRegressor */
def getPredictionsAndLabelsDT(testCV: RDD[Array[String]], model: DecisionTreeRegressionModel): DataFrame = {
    // Hacer predicciones.
    var predictionAndLabels = model.transform(getDataFrame(testCV))
    // Obtener la label y la prediction de los ejemplos de prueba
    predictionAndLabels.select("prediction", "label")

}

/* Función que devuelve el DataFrame con las predicciones y las labels de un modelo RandomForestRegressor */
def getPredictionsAndLabelsRF(testCV: RDD[Array[String]], model: RandomForestRegressionModel): DataFrame = {
  // Hacer predicciones.
  var predictionAndLabels = model.transform(getDataFrame(testCV))
  // Obtener la label y la prediction de los ejemplos de prueba
  predictionAndLabels.select("prediction", "label")

}
/* Función que devuelve el RDD con las predicciones y las labels de un modelo GBTRegressor */
def getPredictionsAndLabelsGBT(testCV: RDD[Array[String]], model: GBTRegressionModel): DataFrame = {
    // Hacer predicciones.
    var predictionAndLabels = model.transform(getDataFrame(testCV))
    // Obtener la label y la prediction de los ejemplos de prueba
    predictionAndLabels.select("prediction", "label")
}

/* Función que obtiene las metricas para la evaluacion */
def getRegressionEvaluator(predictionsAndLabels: DataFrame): (Double, Double, Double, Double) = {
    
    var rmse = new RegressionEvaluator().setMetricName("rmse").evaluate(predictionsAndLabels);
    var mse = new RegressionEvaluator().setMetricName("mse").evaluate(predictionsAndLabels);
    var mae = new RegressionEvaluator().setMetricName("mae").evaluate(predictionsAndLabels);
    var r2 = new RegressionEvaluator().setMetricName("r2").evaluate(predictionsAndLabels);

    (rmse, mse, mae, r2)
}

def crossValidation(subconjuntos: Array[RDD[Array[String]]], modelName:String, init:Int, end:Int): (Double, Int) = {
  // Array que almacena, para cada valor de maxProf, las medias de las tasas de error de la validación cruzada de los subconjuntos
  var meanR2: ArrayBuffer[Double] = ArrayBuffer[Double]()
  var param = 1
  var k = 0

  // Utilizamos Validación cruzada para determinar la características optimas de cada modelo
  for (param <- init to end) {

    // Array que almacena la tasa de error de cada iteración de la validación cruzada
    var array_rmse: ArrayBuffer[Double] = ArrayBuffer[Double]()
    // Array que almacena la aucROC de cada iteración de la validación cruzada
    var array_mse: ArrayBuffer[Double] = ArrayBuffer[Double]()
    // Array que almacena la desviacion estandar de cada iteración de la validación cruzada
    var array_mae: ArrayBuffer[Double] = ArrayBuffer[Double]()
    // Array que almacena los minimos del intervalo de cada iteración de la validación cruzada
    var array_r2: ArrayBuffer[Double] = ArrayBuffer[Double]()
    

    // Utilizamos cada subconjunto como prueba
    for (k <- 0 to subconjuntos.length-1){

      //Obtenemos los subconjuntos
      var (testCV,trainCV) = getSubconjuntos(subconjuntos, k)
      // Creamos un RDD de LabeledPoints
      //var labeledPointsTrain = getRDDLabeledPoint(trainCV)

      var predictionsAndLabels = Seq.empty[(Double, Double)].toDF("k", "v")


      if (modelName == DECISION_TREE){

        // Inducimos un Arbol de Decision
        var model = new DecisionTreeRegressor().
            setLabelCol("label").
            setFeaturesCol("features").
            setMaxDepth(param). 
            setMaxBins(MAXBINS). 
            fit(getDataFrame(trainCV))

        // Obtenemos las predicciones y las labels
        predictionsAndLabels = getPredictionsAndLabelsDT(testCV, model)


      } else if (modelName == RANDOM_FOREST){

        var model = new RandomForestRegressor().
            setLabelCol("label").
            setFeaturesCol("features").
            setMaxDepth(param).  
            setMaxBins(MAXBINS).
            setNumTrees(NUM_TREES).
            fit(getDataFrame(trainCV))

        // Obtenemos las predicciones y las labels
        predictionsAndLabels = getPredictionsAndLabelsRF(testCV, model)


      } else if (modelName == GBT){

        // Inducimos un Arbol de Decision
        var model = new GBTRegressor().
            setLabelCol("label").
            setFeaturesCol("features").
            setMaxIter(param).
            setMaxBins(MAXBINS).
            fit(getDataFrame(trainCV))

        // Obtenemos las predicciones y las labels
        predictionsAndLabels = getPredictionsAndLabelsGBT(testCV, model)

      }

      /********************** TEST *********************/
      // Obtenemos las métricas
        var (rmse, mse, mae, r2) = getRegressionEvaluator(predictionsAndLabels)

        println("-----" + modelName + "-------")
        println("Root Mean Squared Error (RMSE) = " + rmse)
        println("Mean Squared Error (MSE) = " + mse)
        println("Mean Absolute Error (MAE) = " + mae)
        println("Coefficient of Determination (R2) = " + r2)
        println()

        // Añadimos las metricas a los Arrays correspondientes
        array_rmse += rmse
        array_mse += mse
        array_mae += mae
        array_r2 += r2
    }

    // Añadimos a los arrays las medias de las métricas
    meanR2 += array_r2.sum/array_r2.length

    println("[" + param + "] Coeficiente de determinación: " + array_r2.sum/array_r2.length)
    println()
  }
  // Obtenemos maximo de las medias de las tasas de errores
    var maxR2 = meanR2.max

    // Comprobamos con que profundidades del árbol hemos conseguido el minimo valor de tasa de error
  var listMaxParam = meanR2.zipWithIndex.filter(_._1 == maxR2).map(_._2)

  // Maxima profundidad optima
  var paramOptimo = init+listMaxParam.head

  println("\nMejor Coeficiente de determinación para "+modelName+" --> " + maxR2 + "  Parámetro óptimo: " + paramOptimo)

  (maxR2, paramOptimo)
}

/***************************************************
**************** LECTURA DE DATOS   ****************
****************************************************/

var (data, dataTest) = readData()
//veo que es lo que hay en este DF
getDataFrame(data).show


/***************************************************
**************** VALIDACION CRUZADA ****************
****************************************************/

// Repartimos los datos de entrenamiento en 10 subconjuntos
//var subconjuntos = data.randomSplit(Array(0.33,0.33,0.34), 2L)
val subconjuntos = data.randomSplit(Array(0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1))

var (maxDT, maxProf) = crossValidation(subconjuntos, DECISION_TREE, 4, 5)
var (maxRF, maxDepth) = crossValidation(subconjuntos, RANDOM_FOREST, 10, 12)
var (maxGTB, maxIter) = crossValidation(subconjuntos, GBT, 20, 25)

println("\nDT --> Mejor coeficiente de determinación: " + maxDT + "  maxProf: " + maxProf)
println("\nRF --> Mejor coeficiente de determinación: " + maxRF + "  maxDepth: " + maxDepth)
println("\nGBT --> Mejor coeficiente de determinación: " + maxGTB + "  maxIter: " + maxIter)


/***************************************************
***************** GUARDAR MODELOS ******************
****************************************************/

var dataDF = getDataFrame(data)      

// Generamos los modelos
var modelDT = new DecisionTreeRegressor().
            setLabelCol("label").setFeaturesCol("features").setMaxDepth(maxProf).setMaxBins(MAXBINS).fit(dataDF)  
var modelRF = new RandomForestRegressor().
            setLabelCol("label").setFeaturesCol("features").setMaxDepth(maxDepth).setMaxBins(MAXBINS).setNumTrees(NUM_TREES).fit(dataDF)
var modelGBT = new GBTRegressor().
            setLabelCol("label").setFeaturesCol("features").setMaxIter(maxIter).setMaxBins(MAXBINS).fit(dataDF)

// Guardar modelos
modelDT.write.overwrite().save("Modelo/" + DECISION_TREE)
modelRF.write.overwrite().save("Modelo/" + RANDOM_FOREST)
modelGBT.write.overwrite().save("Modelo/" + GBT)

